/*
 * jGuild Project: jRPM
 * Released under the Apache License ( http://www.apache.org/LICENSE )
 */
package com.jguild.jrpm.io;

import java.io.DataInputStream;
import java.io.IOException;

import org.apache.log4j.Logger;

import com.jguild.jrpm.io.constant.RPMIndexType;

/**
 * RPM Index Entry (16 byte).
 * 
 * <ol>
 * <li>tag; what the data is for (4 bytes)</li>
 * <li>type; what type of data is stored (see RPMFileType.java) (4 bytes)
 * </li>
 * <li>offset; where to the data begins (4 bytes)</li>
 * <li>count; how much data of the type is there (4 bytes)</li>
 * </ol>
 * 
 * @version $Id: IndexEntry.java,v 1.5 2004/05/06 20:59:22 mkuss Exp $
 */
public class IndexEntry {

    private static final int INDEX_LENGTH = 16;

    private static final Logger logger = Logger.getLogger(IndexEntry.class);

    private RPMIndexType type;

    private long count;

    private long offset;

    private long tag;

    /**
     * Constructs a IndexEntry out of a InputStream. This will read the tag id,
     * the data type, the offset of the data and the number of elements of
     * data. Note that the number of elements varies in its meaning depending
     * on the data type.
     * 
     * @param inputStream
     *           An InputStream containing a rpm file.
     * 
     * @throws IOException
     *            If an error occurs during reading from stream
     */
    public IndexEntry(DataInputStream inputStream) throws IOException {
        tag = inputStream.readInt();

        if (logger.isDebugEnabled()) {
            logger.debug("tag:" + tag);
        }

        type = RPMIndexType.getRPMIndexType(inputStream.readInt());

        if (logger.isDebugEnabled()) {
            logger.debug("type: " + type);
        }

        offset = inputStream.readInt();

        if (logger.isDebugEnabled()) {
            logger.debug("offset: " + offset);
        }

        count = inputStream.readInt();

        if (logger.isDebugEnabled()) {
            logger.debug("count: " + count);
        }

        check(!type.equals(RPMIndexType.UNKNOWN));
    }

    /**
     * Get the number of elements of this data type
     * 
     * @return The number of elements
     */
    public long getCount() {
        return count;
    }

    /**
     * Returns the offset of this data type
     * 
     * @return The offset
     */
    public long getOffset() {
        return offset;
    }

    /**
     * Get the size of this index entry in bytes
     * 
     * @return The size
     */
    public int getSize() {
        return INDEX_LENGTH;
    }

    /**
     * Get the tag id as a long
     * 
     * @return The tag id
     */
    public long getTag() {
        return tag;
    }

    /**
     * Get the data type of this entry
     * 
     * @return The data type
     */
    public RPMIndexType getType() {
        return type;
    }

    /**
     * Asserts a boolean value and throws an exception if it is false
     * 
     * @param test
     *           A boolean test variable
     * 
     * @throws IOException
     *            if the variable test is false
     */
    private static final void check(boolean test) throws IOException {
        if (!test) { throw new IOException("Corrupted archive"); }
    }
}
